# This file contains the Jot "engine" for authoring markup.
# 

#### TODO:
# 
# - Import util dependencies.
# - Import scope dependency.
# - Move Node::toString implementation into subclasses

# Require scope
{scope, COFFEESCRIPT_HELPERS } = require './scope'

#### Helpers

# isFunction, isArray, isString are borrowed from underscorejs
isFunction = (x) -> !!(x and x.constructor and x.call and x.apply)
isString = (x) -> !!(x is '' or (x and x.charCodeAt and x.substr))
isArray = (x) -> x and x.slice and x.join
isNumber = (x) -> x? and x.toPrecision and x.toExponential


isFn = (x) -> isFunction(x) or isString(x) and !!x.match /^\s*\(?function\b/

# Escape HTML entities
escapeXml = (x) ->
  x = x.replace /&/g, '&amp;'
  x = x.replace />/g, '&gt;'
  x = x.replace /</g, '&lt;'

# Remove values from an array. Modifies the array.
remove = (arr, vals...) -> 
  i = arr.length; 
  while i-- 
    arr.splice i, 1 if arr[i] in vals 

# jQuery's extend: copy attributes from one object to another.
extend = (x, more...) -> x[k] = v for k,v of o for o in more; x
    
# Given a level of indentation, return the whitespace
# needed to indent.
indent = (level) -> Array(level+1).join INDENT

# The default indent is two spaces
INDENT = '  '

# ###Class Node
# A node in the markup tree.
exports.Node = class Node
  newline: (format, level) -> if format then "\n#{indent level}" else ''
    

exports.Element = class Element extends Node
  # Constructor takes the tagname
  constructor: (@tag) ->
    @attrs = {}
    @children = []
    
  # Get or set this node's tagname
  tagname: (newtag) -> 
    if newtag then @tag = newtag; @
    else @tag
  
  # Adds children to this node
  append: (childs...) -> @children.push c for c in childs; @

  # Sets attributes on this node
  attr: (k, v) -> 
    if isString k then @attrs[k] = v
    else @attr k2, v for k2, v of k
    @

  # Returns an xml-like string representation of this node.
  toString: (format, level = 0) ->
    @streamString arr = [], format, level
    arr.join ''
    
  # Use an array like a stream for efficient
  # string append.
  streamString: (stream, format, level) ->
    end = @newline format, level
    s = (s) -> stream.push s
    
    s "#{end}<#{@tag}"
    s "#{(" #{k}=\"#{v}\"" for k, v of @attrs).join ''}"
    if not @requiresContent() and @children.length is 0
      s '/>'
    else 
      s '>'
      format = false if (@tag is 'textarea') or (@tag is 'pre') or @attrs.contenteditable
      for c in @children
        c.streamString?(stream, format, level + 1) or
          s c.toString format, level + 1 
        if c instanceof Element or c instanceof Cdata then heavy = true
      s end if heavy
      s "</#{@tag}>"
    true

  requiresContent: -> false
  
QUOTE_CHAR = '`'

exports.TextNode = class TextNode extends Node
  REG_NOESCAPE = ///^\s*#{QUOTE_CHAR}([\s\S]*)/// # A backtick means "do not escape text"
  constructor: (@text = '', @escape = true) ->
    throw 'not a string' if typeof @text isnt 'string'
    if @escape and m = @text.match REG_NOESCAPE
      @escape = false
      @text = m[1]
  indented: (format) -> 
    if format then @text.replace /\n/g, @newline arguments...
    else @text
    
  toString: -> 
    text = @indented arguments...
    if @escape then escapeXml text else text

exports.Cdata = class Cdata extends TextNode
  constructor: (@text) ->
  toString: (f, level) ->
    text = @indented f, level + 1
    @newline(arguments...) + @end[0] + text + @end[1]
  end: ['<![CDATA[', ']]>']
  

exports.Comment = class Comment extends Cdata
  end: ['<!--', '-->']
  

exports.Doctype = class Doctype extends Node
  constructor: (@type) ->
  toString: -> "\n#{types[@type]}"

  # Borrowed from CoffeeKup coffeekup.org
  types =
    html5: '<!DOCTYPE html>'
    xml: '<?xml version="1.0" encoding="utf-8" ?>'
    default: '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
    transitional: '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'
    strict: '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'
    frameset: '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Frameset//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd">'
    '1.1': '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',
    basic: '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML Basic 1.1//EN" "http://www.w3.org/TR/xhtml-basic/xhtml-basic11.dtd">'
    mobile: '<!DOCTYPE html PUBLIC "-//WAPFORUM//DTD XHTML Mobile 1.2//EN" "http://www.openmobilealliance.org/tech/DTD/xhtml-mobile12.dtd">'
  
  
    
# ###Class Content
# A context for building markup fragments.
# All tags and macros added to this context become 
# instance properties that can be called outside 
# of a scoped function (for convenience).
exports.Context = class Context
  
  
  constructor: ({@tags}) ->
    @tags ?= []
    @format = true
    
    # Keeps track of various aspects of the context
    # While tags and macros are being called.
    @state =  
      # The current value of 'this' inside a macro or mixin.
      self: null
      # The current node being operated on.
      node: null
      # The current list of nodes without parents.
      orphans: []
    
    # Keep track of scoped functions
    @scoped = []
    # Setup the initial local scope    
    @locals = {}
    # Declare all tags
    @declareTag tag for tag in @tags
    # Add CoffeeScript helpers in if macros
    # or scoped functions are written in coffee-script
    @local COFFEESCRIPT_HELPERS

    # Other local functions
    @local
      node: => @state.node
      scope: => @scope arguments...
      cdata: @cdata
      xml: @xml
      text: @text
      cdata: @cdata
      note: @comment
      call: @call
      macro: @flexargs true, (locals, fn) -> @inlineMacro fn
    
    @delegateToNode 'append', 'attr', 'tagname'
    
    @alias text: 'append'
  
  xml: (txt) => @text txt, false
  text: (txt, escape = true) => @orphan new TextNode txt, escape
  cdata: (txt) => @orphan new Cdata txt
  comment: (txt) => @orphan new Comment txt
  
  # Push a state change, call f, then pop the change
  withState: (o, f) ->
    # push state
    old = {}; for k of o
      old[k] = @state[k]; @state[k] = o[k]
      
    ret = f()
    
    # pop state
    @state[k] = old[k] for k of o
    
    ret
  
  # functions that delegate to the current node
  # made available in lexical scope
  delegateToNode: (names...) -> 
    for name in names then do (name) =>
      @local name, @[name] = => @state.node[name] arguments...
    
  
  # Public functions for modifying the local scope.

  # Add variable(s) to local scope.
  local: (nameOrHash, value) -> 
    if isString nameOrHash 
      if value? then @locals[nameOrHash] = value
      else @locals[nameOrHash]
    else @local k, v for k, v of nameOrHash
  
  
  # Remove variable(s) from local scope
  removeLocal: (names...) -> delete @local[name] for name in names
  
  # Alias a variable in local scope
  alias: (o) -> 
    @local name, @local dest for name, dest of o
  
  orphan: (n) -> @state.orphans.push n; n
  unorphan: (n) -> 
    @state.orphans = 
      (orph for orph in @state.orphans when n isnt orph); n 
      
  # create a tag with the given name and args
  # This is a helper function.
  # Assumes all function arguments are compiled.
  callTag: (tag, args) =>
    @nodeCall n = new @Element(tag), args
    # Add new node to orphans until we 
    # can find it a parent.
    n.call = => @nodeCall n, arguments
    @orphan n
  
  # Temporary
  declare: (names...) ->
    @local name, 0 for name in names
  
  # Add a simple tag function to the local scope
  declareTag: (tag, tagname = tag) ->
    @local tag, => @callTag tagname, arguments 
    @[tag] = (args...) => 
      @callTag tagname,
        isFn(x) and @scope(x) or x for x in args
      @render [@state.orphans.pop()]
  
  

  # The Node class to instantiate
  Element: Element
  
  # Call fn with current state.self as 'this'
  # Append any emitted orphans to current node.
  mixin: (fn) ->
    ret = null
    nodes = @trapOrphans true, => ret = fn.call @state.self
    # template
    if isString ret then @processArgString ret
    # mixin
    else @append nodes...
    
  # Process an argument passed to a tag call
  processArg: (arg) ->
    arg = arg.toString() if isNumber arg
  
    # The case when an emitting function is called in an argument
    # to this node.
    if isArray arg then @processArg a for a in arg
      
    else if (node = arg) instanceof Node
      @append @unorphan node
      
    # Call the function and append the emitted nodes
    # f is already scoped.
    else if isFunction arg then @mixin arg
      
    else if isString arg then @processArgString arg

    else @attr arg

  # Process a string argument
  processArgString: (s) ->  @append new TextNode s

  
  nodeCall: (n, args) ->
    @withState node: n, => @call args...
    n
  
  call: (args...) => @processArg arg for arg in args
  
  trapOrphans: (remove, f) ->
    len = @state.orphans.length
    f()
    ret = @state.orphans[len..]
    @state.orphans = @state.orphans[...len] if remove
    ret
  
  # The external representation of an array of nodes. 
  # The default is to stringify them
  render: (nodes) -> @stringify nodes
  
  # Return a single string of all nodes combined.
  stringify: (nodes) -> 
    str = (n.toString @format for n in nodes).join('')
    if str[0] is '\n' then str[1..] else str
  
  
  # Allow flexible calling of function
  # If a name is detected, result will be 
  # added to local scope
  # obj arguments are is a hash<name:fn>
  # Options:
  # - (name, fn)
  # - (name, locals, fn)
  # - (fn)
  # - (locals, fn)
  # - (fn, locals)
  # - (locals, obj)
  # - (obj)
  @flexargs = @::flexargs = (needsRender, impl) ->
    externify = (intern) ->
      if needsRender
        (args...) => 
          @render @trapOrphans true, -> 
            intern.apply null, args
      else intern
    
    f = (name, locals, fn) ->

      # case: (name, locals, fn) or (name, fn)
      if isString(name) and not isFn name
        [locals, fn] = [null, locals] if not fn
        
        intern = @local name, impl.call @, locals, fn
        intern._name = name
        @[name] = externify.call @, intern
          
      # Eliminate possibility of obj
      # case: (locals, fn) or (fn, locals)
      else if isFn(name) or b = isFn locals
        [locals, fn] = 
          if b 
            [name, locals] 
          else 
            [locals, name]
            
        intern = impl.call @, locals, fn
        externify.call @, intern
      
      # We are dealing with an obj
      else 
        
        # case: (locals, obj)
        if (obj = locals) then locals = name
        
        # case: (obj)
        else obj = name
        
        f.call @, name, locals, fn for name, fn of obj
  
  scope: 
    @flexargs false, (locals, fn) -> 
      fn = scope extend({}, @locals, locals), fn 
      fn.moreLocals = locals
      @scoped.push fn; {state} = @
      (args...) -> fn.apply state.self, args
        
      
  
  # The value of 'this' is the first argument
  # passed to the macro when called.
  # macros look and behave like tags, only they
  # generate more than a single tag of content.
  # Macros get wrapped and return an array of top-level nodes
  # created in its scope.
  macro: 
    @flexargs true, (locals, fn) -> 
      @inlineMacro @scope locals, fn; 
      
  # an 'inline' function is one that is defined inside
  # a scoped function.
  inlineMacro: (fn) ->
    {state} = @
    (args...) => 
      call = -> fn.apply state.self, args
      @MacroResult \
        @trapOrphans false, => 
          if args.length then @withState self: args[0], call
          else call()
 

  MacroResult: (nodes) ->
    nodes.call = => @nodeCall n, arguments for n in nodes; nodes
    nodes.toString = => @stringify nodes
    nodes

  
  # if a macro expands to a list of nodes,
  # a template expands to a string.
  # It's like an inside-out macro: the nodes are 
  # embedded in the string, rather than vice-versa.
  template: 
    @flexargs false, (locals, fn) ->
      @inlineTemplate @scope locals, fn; 
      
  
  inlineTemplate: (fn) ->
    {state} = @
    (args...) =>
      str = null
      call = -> str = fn.apply state.self, args
      # Remove orphans... they are embedded in the string
      @trapOrphans true, =>
        if args.length then @withState self: args[0], call
        else call()
      str

  # Update scoped functions with current locals
  rescope: ->
    for fn in @scoped 
      fn.scope extend({}, @locals, fn.moreLocals)
          
  # Compile and call the function
  run: (locals, fn, args...) ->
    if isFn locals
      args.unshift fn; fn = locals; locals = null
    @macro(locals, fn) args...

# !!!You only need MACRO-style wrapper (trapOrphans, etc) if you are returning more than one NODE!!!!!!
# !!!Otherwise you can just use scope!!!








# An HTML Context

ADD_DEFAULT_CSS_CLASS = true

htmlRequiresContent = do ->
  o = {}
  o[tag] = true for tag in "script textarea div h2 p h1".split ' '
  o

exports.HtmlElement = class HtmlElement extends Element
  requiresContent: -> htmlRequiresContent[@tag]
    

exports.HtmlContext = class HtmlContext extends Context
  # mined from http://www.w3schools.com/html5/html5_reference.asp
  HTMLTAGS = "a,abbr,acronym,address,applet,area,article,aside,audio,b,base,basefont,bdo,big,blockquote,body,br,button,canvas,caption,center,cite,code,col,colgroup,command,datalist,dd,del,details,dfn,dir,div,dl,dt,em,embed,fieldset,figcaption,figure,font,footer,form,frame,frameset,h1,h2,h3,h4,h5,h6,head,header,hgroup,hr,html,i,iframe,img,input,ins,keygen,kbd,label,legend,li,link,map,mark,menu,meta,meter,nav,noframes,noscript,object,ol,optgroup,option,output,p,param,pre,progress,q,rp,rt,ruby,s,samp,script,section,select,small,source,span,strike,strong,style,sub,summary,sup,table,tbody,td,textarea,tfoot,th,thead,time,title,tr,tt,u,ul,video,wbr,xmp".split ','
  
  REG_CLASS = /^\.-?[_a-zA-Z][_a-zA-Z0-9-]*/
  REG_ID = /^#([A-Za-z][A-Za-z0-9\-:\._]*)$/
  
  constructor: -> 
    super tags: HTMLTAGS 
    @alias 
      _: 'div'
      o: 'div'
      
    @scope
      #nav: -> _ '.nav', arguments...
      javascript: -> script {type:'text/javascript'}, arguments...
      css: (href, media = "screen, projection") -> 
        link {href, media, rel:"stylesheet", type:"text/css"}
    showdown = require('./../vendor/showdown').Showdown
    @local
      $: (s) => @text "\n#{showdown.makeHtml(s)}", false
      addClass: => @addClass arguments...
      doctype: (type) => @orphan new Doctype type
    @declareTag 'var_', 'var'
  
  Element: HtmlElement
  
  processArgString: (s) ->
    if s.match REG_CLASS then @addClass s.split('.').join ' '
    else if m = s.match REG_ID then @attr 'id', m[1]
    else super s
  
  addClass: (c) -> 
    n = @state.node
    cls = (n.attrs.class or '') + ' ' + c
    n.attrs.class = (c for c in cls.split ' ' when c).join ' '
    null
  
  markdown: (s) -> @locals.$(s).toString()
 
